package ru.dbotthepony.kstarbound.world.entities

import com.google.gson.JsonObject
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.defs.EntityDamageTeam
import ru.dbotthepony.kstarbound.defs.actor.HumanoidData
import ru.dbotthepony.kstarbound.defs.actor.HumanoidEmote
import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import ru.dbotthepony.kstarbound.math.Interpolator
import ru.dbotthepony.kstarbound.network.syncher.GroupElement
import ru.dbotthepony.kstarbound.network.syncher.MasterElement
import ru.dbotthepony.kstarbound.network.syncher.networkedBoolean
import ru.dbotthepony.kstarbound.network.syncher.networkedData
import ru.dbotthepony.kstarbound.network.syncher.networkedEnum
import ru.dbotthepony.kstarbound.network.syncher.networkedEnumExtraStupid
import ru.dbotthepony.kstarbound.network.syncher.networkedEnumStupid
import ru.dbotthepony.kstarbound.network.syncher.networkedEventCounter
import ru.dbotthepony.kstarbound.network.syncher.networkedFixedPoint
import ru.dbotthepony.kstarbound.network.syncher.networkedString
import java.io.DataOutputStream
import java.util.UUID
import kotlin.properties.Delegates

class PlayerEntity() : HumanoidActorEntity("/") {
	enum class State(override val jsonName: String) : IStringSerializable {
		IDLE("Idle"),
		WALK("Walk"),
		RUN("Run"),
		JUMP("Jump"),
		FALL("Fall"),
		SWIM("Swim"),
		SWIM_IDLE("SwimIdle"),
		TELEPORT_IN("TeleportIn"),
		TELEPORT_OUT("TeleportOut"),
		CROUCH("Crouch"),
		LOUNGE("Lounge");
	}

	override fun writeToNetwork(stream: DataOutputStream, isLegacy: Boolean) {
		TODO("Not yet implemented")
	}

	val networkGroup = MasterElement(GroupElement())

	var state by networkGroup.upstream.add(networkedEnum(State.IDLE))
	var shifting by networkGroup.upstream.add(networkedBoolean())
	private var xAimPosition by networkGroup.upstream.add(networkedFixedPoint(0.003125))
	private var yAimPosition by networkGroup.upstream.add(networkedFixedPoint(0.003125).also { it.interpolator = Interpolator.Linear })
	var humanoidData by networkGroup.upstream.add(networkedData(HumanoidData(), HumanoidData.CODEC, HumanoidData.LEGACY_CODEC))
	var teamState by networkGroup.upstream.add(networkedData(EntityDamageTeam(), EntityDamageTeam.CODEC, EntityDamageTeam.LEGACY_CODEC))
	val landed = networkGroup.upstream.add(networkedEventCounter())
	var chatMessage by networkGroup.upstream.add(networkedString())
	val newChatMessage = networkGroup.upstream.add(networkedEventCounter())
	var emote by networkGroup.upstream.add(networkedEnumExtraStupid(HumanoidEmote.IDLE))

	override val aimPosition: Vector2d
		get() = Vector2d(xAimPosition, yAimPosition)

	override val isApplicableForUnloading: Boolean
		get() = false

	override fun defs(): Collection<JsonObject> {
		return emptyList()
	}

	var uuid: UUID by Delegates.notNull()
		private set
}
